const coral = @import("coral");

const tokens = @import("./tokens.zig");

name: []const coral.io.Byte,
allocator: coral.io.Allocator,
arena: coral.arena.Stacking,
statements: Statement.List,
error_buffer: coral.list.ByteStack,

pub const Expression = union (enum) {
	nil_literal,
	true_literal,
	false_literal,
	number_literal: []const coral.io.Byte,
	string_literal: []const coral.io.Byte,
	table_literal: NamedList,
	grouped_expression: *Expression,
	get_local: []const coral.io.Byte,

	call_system: struct {
		identifier: []const coral.io.Byte,
		argument_expressions: List,
	},

	binary_operation: struct {
		operator: BinaryOperator,
		lhs_expression: *Expression,
		rhs_expression: *Expression,
	},

	unary_operation: struct {
		operator: UnaryOperator,
		expression: *Expression,
	},

	pub const BinaryOperator = enum {
		addition,
		subtraction,
		multiplication,
		divsion,
		equals_comparison,
		greater_than_comparison,
		greater_equals_comparison,
		less_than_comparison,
		less_equals_comparison,

		fn token(self: BinaryOperator) tokens.Token {
			return switch (self) {
				.addition => .symbol_plus,
				.subtraction => .symbol_minus,
				.multiplication => .symbol_asterisk,
				.divsion => .symbol_forward_slash,
				.equals_comparison => .symbol_double_equals,
				.greater_than_comparison => .symbol_greater_than,
				.greater_equals_comparison => .symbol_greater_equals,
				.less_than_comparison => .symbol_less_than,
				.less_equals_comparison => .symbol_less_equals,
			};
		}
	};

	pub const NamedList = coral.list.Stack(struct {
		identifier: []const coral.io.Byte,
		expression: Expression,
	});

	pub const List = coral.list.Stack(Expression);
};

const ExpressionParser = fn (self: *Self, tokenizer: *tokens.Tokenizer) ParseError!Expression;

pub const ParseError = error {
	OutOfMemory,
	BadSyntax,
};

const Self = @This();

pub const Statement = union (enum) {
	return_expression: Expression,
	return_nothing,

	set_local: struct {
		identifier: []const coral.io.Byte,
		expression: Expression,
	},

	call_system: struct {
		identifier: []const coral.io.Byte,
		argument_expressions: Expression.List,
	},

	const List = coral.list.Stack(Statement);
};

const UnaryOperator = enum {
	boolean_negation,
	numeric_negation,
};

fn binary_operation_parser(
	comptime parse_next: ExpressionParser,
	comptime operators: []const Expression.BinaryOperator) ExpressionParser {

	const BinaryOperationParser = struct {
		fn parse(self: *Self, tokenizer: *tokens.Tokenizer) ParseError!Expression {
			const allocator = self.arena.as_allocator();
			var expression = try parse_next(self, tokenizer);

			inline for (operators) |operator| {
				const token = comptime operator.token();

				if (tokenizer.is_token(coral.io.tag_of(token))) {
					tokenizer.step();

					if (tokenizer.token == null) {
						return self.report(tokenizer, "expected other half of expression after `" ++ comptime token.text() ++ "`");
					}

					expression = .{
						.binary_operation = .{
							.operator = operator,
							.lhs_expression = try coral.io.allocate_one(allocator, expression),
							.rhs_expression = try coral.io.allocate_one(allocator, try parse_next(self, tokenizer)),
						},
					};
				}
			}

			return expression;
		}
	};

	return BinaryOperationParser.parse;
}

pub fn error_message(self: Self) []const coral.io.Byte {
	return self.error_buffer.values;
}

pub fn free(self: *Self) void {
	self.arena.free();
	self.statements.free();
	self.error_buffer.free();
}

pub fn make(allocator: coral.io.Allocator, ast_name: []const coral.io.Byte) Self {
	return Self{
		.arena = coral.arena.Stacking.make(allocator, 4096),
		.error_buffer = coral.list.ByteStack.make(allocator),
		.statements = Statement.List.make(allocator),
		.allocator = allocator,
		.name = ast_name,
	};
}

fn report(self: *Self, tokenizer: *tokens.Tokenizer, message: []const coral.io.Byte) ParseError {
	coral.utf8.print_formatted(coral.list.stack_as_writer(&self.error_buffer), "{name}@{line}: {message}", .{
		.name = self.name,
		.line = tokenizer.lines_stepped,
		.message = message,
	}) catch return error.OutOfMemory;

	return error.BadSyntax;
}

pub fn list_statements(self: Self) []const Statement {
	return self.statements.values;
}

pub fn parse(self: *Self, tokenizer: *tokens.Tokenizer) ParseError!void {
	self.free();

	const allocator = self.arena.as_allocator();
	var has_returned = false;

	while (true) {
		tokenizer.skip(.newline);

		switch (tokenizer.token orelse return) {
			.keyword_return => {
				if (has_returned) {
					return self.report(tokenizer, "multiple returns in function scope but expected only one");
				}

				try self.statements.push_one(get_statement: {
					tokenizer.step();

					if (!tokenizer.is_token_null_or(.newline)) {
						break: get_statement .{.return_expression = try self.parse_expression(tokenizer)};
					}

					if (!tokenizer.is_token_null_or(.newline)) {
						return self.report(tokenizer, "unexpected token after return");
					}

					break: get_statement .return_nothing;
				});

				has_returned = true;
			},

			.identifier => |identifier| {
				tokenizer.step();

				const no_effect_message = "statement has no effect";

				switch (tokenizer.token orelse return self.report(tokenizer, no_effect_message)) {
					.newline => return self.report(tokenizer, no_effect_message),

					.symbol_equals => {
						tokenizer.step();

						if (tokenizer.token == null) {
							return self.report(tokenizer, "expected expression after `=`");
						}

						try self.statements.push_one(.{
							.set_local = .{
								.expression = try self.parse_expression(tokenizer),
								.identifier = identifier,
							},
						});

						if (!tokenizer.is_token_null_or(.newline)) {
							return self.report(tokenizer, "unexpected token after assignment");
						}
					},

					else => return self.report(tokenizer, "expected `=` after local"),
				}
			},

			.special_identifier => |identifier| {
				tokenizer.step();

				const missing_arguments_message = "system call is missing arguments";

				switch (tokenizer.token orelse return self.report(tokenizer, missing_arguments_message)) {
					.newline => return self.report(tokenizer, missing_arguments_message),

					.symbol_paren_left => {
						tokenizer.step();

						var expressions_list = Expression.List.make(allocator);

						while (true) {
							if (tokenizer.is_token(.symbol_paren_right)) {
								break;
							}

							try expressions_list.push_one(try self.parse_expression(tokenizer));

							switch (tokenizer.token orelse return self.report(tokenizer, "unexpected end after after `(`")) {
								.symbol_comma => continue,
								.symbol_paren_right => break,
								else => return self.report(tokenizer, "expected `)` or argument after `(`"),
							}
						}

						tokenizer.step();

						try self.statements.push_one(.{
							.call_system = .{
								.argument_expressions = expressions_list,
								.identifier = identifier,
							},
						});
					},

					else => return self.report(tokenizer, "expected `=` after local"),
				}
			},

			else => return self.report(tokenizer, "invalid statement"),
		}
	}
}

const parse_comparison = binary_operation_parser(parse_term, &.{
	.greater_than_comparison,
	.greater_equals_comparison,
	.less_than_comparison,
	.less_equals_comparison
});

const parse_equality = binary_operation_parser(parse_comparison, &.{
	.equals_comparison,
});

const parse_expression = binary_operation_parser(parse_equality, &.{
	.addition,
	.subtraction,
});

fn parse_factor(self: *Self, tokenizer: *tokens.Tokenizer) ParseError!Expression {
	const allocator = self.arena.as_allocator();

	switch (tokenizer.token orelse return self.report(tokenizer, "expected operand after operator")) {
		.symbol_paren_left => {
			tokenizer.skip(.newline);

			if (tokenizer.token == null) {
				return self.report(tokenizer, "expected an expression after `(`");
			}

			const expression = try self.parse_expression(tokenizer);

			if (!tokenizer.is_token(.symbol_paren_right)) {
				return self.report(tokenizer, "expected a closing `)` after expression");
			}

			tokenizer.step();

			return Expression{.grouped_expression = try coral.io.allocate_one(allocator, expression)};
		},

		.keyword_nil => {
			tokenizer.step();

			return .nil_literal;
		},

		.keyword_true => {
			tokenizer.step();

			return .true_literal;
		},

		.keyword_false => {
			tokenizer.step();

			return .false_literal;
		},

		.number => |value| {
			tokenizer.step();

			return Expression{.number_literal = value};
		},

		.string => |value| {
			tokenizer.step();

			return Expression{.string_literal = value};
		},

		.special_identifier => |identifier| {
			tokenizer.skip(.newline);

			var expression_list = Expression.List.make(allocator);

			while (true) {
				switch (tokenizer.token orelse return self.report(tokenizer, "expected expression or `)` after `(`")) {
					.symbol_paren_right => {
						tokenizer.step();

						return Expression{
							.call_system = .{
								.identifier = identifier,
								.argument_expressions = expression_list,
							},
						};
					},

					else => {
						try expression_list.push_one(try self.parse_expression(tokenizer));

						switch (tokenizer.token orelse return self.report(tokenizer, "expected `,` or `)` after argument")) {
							.symbol_comma => continue,

							.symbol_paren_right => {
								tokenizer.step();

								return Expression{
									.call_system = .{
										.identifier = identifier,
										.argument_expressions = expression_list,
									},
								};
							},

							else => return self.report(tokenizer, "expected `,` or `)` after argument"),
						}
					},
				}
			}
		},

		.identifier => |identifier| {
			tokenizer.step();

			return Expression{.get_local = identifier};
		},

		.symbol_brace_left => {
			var table_fields = Expression.NamedList.make(allocator);

			tokenizer.skip(.newline);

			while (true) {
				switch (tokenizer.token orelse return self.report(tokenizer, "unexpected end of table literal")) {
					.symbol_brace_right => {
						tokenizer.step();

						return Expression{.table_literal = table_fields};
					},

					.identifier => |identifier| {
						tokenizer.skip(.newline);

						if (!tokenizer.is_token(.symbol_equals)) {
							return self.report(tokenizer, "expected `=` after identifier");
						}

						tokenizer.skip(.newline);

						if (tokenizer.token == null) {
							return self.report(tokenizer, "unexpected end after `=`");
						}

						try table_fields.push_one(.{
							.expression = try self.parse_expression(tokenizer),
							.identifier = identifier,
						});

						switch (tokenizer.token orelse return self.report(tokenizer, "unexpected end of table")) {
							.symbol_comma => tokenizer.skip(.newline),

							.symbol_brace_right => {
								tokenizer.step();

								return Expression{.table_literal = table_fields};
							},

							else => return self.report(tokenizer, "expected `,` or `}` after expression"),
						}
					},

					else => return self.report(tokenizer, "expected `}` or fields in table literal"),
				}
			}
		},

		.symbol_minus => {
			tokenizer.skip(.newline);

			if (tokenizer.token == null) {
				return self.report(tokenizer, "expected expression after numeric negation (`-`)");
			}

			return Expression{
				.unary_operation = .{
					.expression = try coral.io.allocate_one(allocator, try self.parse_factor(tokenizer)),
					.operator = .numeric_negation,
				},
			};
		},

		.symbol_bang => {
			tokenizer.skip(.newline);

			if (tokenizer.token == null) {
				return self.report(tokenizer, "expected expression after boolean negation (`!`)");
			}

			return Expression{
				.unary_operation = .{
					.expression = try coral.io.allocate_one(allocator, try self.parse_factor(tokenizer)),
					.operator = .boolean_negation,
				},
			};
		},

		else => return self.report(tokenizer, "unexpected token in expression"),
	}
}

const parse_term = binary_operation_parser(parse_factor, &.{
	.multiplication,
	.divsion,
});
